import { IRenderer } from "./render";
import Hist from "redoundo";
/** Interface of a parental `Actions` object that keeps information which
 *  is necessary to initialize `SubActions` object.
 */
export interface IActionsParent {
    /** Renderer that should be called by any `Actions` object for page re-rendering. */
    readonly Renderer: IRenderer;
    /** Object of `redoundo.Hist` class is needed for redo/undo functionality. */
    readonly History: Hist;
}
/** Class of `hyperoop` top-level actions. When you properly set new values
 *  to entries of the `State` property then the corresponding DOM element will
 *  be redrawn automatically. The correspondence between `Actions` object
 *  and this DOM element is established by `init` function call.
 */
export declare class Actions<S extends {}> {
    /** This property, along with the `Remember` property, is used to change
     *  the state of an application and automatically redraw the user interface.
     */
    readonly State: S;
    /** This property has almost the same functionality as the `State` property,
     *  with the difference that changes made to it can be undone by calling the
     *  `History.undo` method.
     */
    readonly Remember: S;
    /** You can force redraw of the user interface by calling the `sheduleRender`
     *  method provided by this property. Usually you do not need to call it directly.
     */
    readonly Renderer: IRenderer;
    /** Object of `redoundo.Hist` class is needed for redo/undo functionality. */
    readonly History: Hist;
    private orig;
    private state;
    private remember;
    private renderer;
    /** Construct an `Action` object, setting the initial `state` to it and optionally describing
     *  the `hist` object of type `redoundo.Hist`. If the `hist` argument is of the `number`
     *  type, the constructor will itself create the `History` object of the given length.
     */
    constructor(start: S, hist?: number | Hist);
    /** The page is redrawn automatically every time you set a new value for any entry of the `State`
     *  property. To set values for several inputs at once with only one automatic redraw, use
     *  this method, passing it a partial description of the new state and the optional parameter
     *  `remember` (`false` by default). If the value of the parameter `remember` is set to `true`, then
     *  the changes can be undone by calling `History.undo` method.
     */
    set(s: Partial<S>, remember?: boolean): void;
    /** This method is used to initialize the `Actions` object by an instance of the `IRenderer` interface.
     *  Usually you do not need to call it directly.
     */
    init(r: IRenderer): void;
}
/** Usually, you only need to manually initialize top-level `Action` objects (by calling the `init` function).
 * This class is used for lower level action-objects that inherit the necessary properties from higher level
 * objects. Thus, you do not need to initialize the `SubActions` of the lower levels manually.
 */
export declare class SubActions<S extends {}> extends Actions<S> {
    /** Construct an `Action` object, setting the initial `state` and parental `Actions`
     *  object. The last will be used for initializing new instance by renderer and history objects.
     */
    constructor(start: S, parent: IActionsParent);
}
