(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (factory((global.hyperoop = {})));
}(this, (function (exports) { 'use strict';

    function isChildren(x) {
        return !x.pop;
    }
    function h(name, attributes) {
        var rest = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            rest[_i - 2] = arguments[_i];
        }
        var ch = [];
        rest.reverse();
        while (rest.length) {
            var nod = rest.pop();
            if (nod && !isChildren(nod)) {
                for (var i = nod.length - 1; i >= 0; i--) {
                    rest.push(nod[i]);
                }
            }
            else if (nod != null && nod !== true && nod !== false) {
                ch.push(nod);
            }
        }
        if (typeof name === "function") {
            return name(attributes || {}, ch);
        }
        var node = {
            attributes: attributes || {},
            children: ch,
            key: attributes && attributes.key,
            nodeName: name,
        };
        return node;
    }

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation. All rights reserved.
    Licensed under the Apache License, Version 2.0 (the "License"); you may not use
    this file except in compliance with the License. You may obtain a copy of the
    License at http://www.apache.org/licenses/LICENSE-2.0

    THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
    WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
    MERCHANTABLITY OR NON-INFRINGEMENT.

    See the Apache Version 2.0 License for specific language governing permissions
    and limitations under the License.
    ***************************************************************************** */
    /* global Reflect, Promise */

    var extendStatics = function(d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };

    function __extends(d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    }

    /**
     * Creates Proxy that calls `after` callback after set or delete entries of a `target`.
     *
     *
     * @param target target object
     * @param after callback to execute after set or delete entries of `target`
     */
    function make(target, after) {
        return new Proxy(target, {
            set: function (t, k, v) {
                if (k in t && t[k] === v) {
                    return true;
                }
                t[k] = v;
                after();
                return true;
            },
            deleteProperty: function (t, k) {
                if (k in t) {
                    delete t[k];
                    after();
                }
                return true;
            },
        });
    }
    /**
     * Creates Proxy that calls `after` callback after set or delete entries of a `target`.
     * Set or delete actions can be (re, un)done using `redoundo.Hist` argument.
     *
     * @param target target object
     * @param after callback to execute after set or delete entries of `target`
     * @param hist `redoundo.Hist` object
     */
    function makeH(target, after, hist) {
        if (!hist) {
            return null;
        }
        return new Proxy(target, {
            set: function (t, k, v) {
                var was = k in target;
                var oldVal = null;
                if (was) {
                    oldVal = target[k];
                    if (oldVal === v) {
                        return true;
                    }
                }
                var redo = function () {
                    target[k] = v;
                    after();
                };
                var undo = function () {
                    if (was) {
                        target[k] = oldVal;
                    }
                    else {
                        delete target[k];
                    }
                    after();
                };
                hist.add({ Redo: redo, Undo: undo });
                return true;
            },
            deleteProperty: function (t, k) {
                if (!(k in t)) {
                    return true;
                }
                var v = t[k];
                var redo = function () {
                    delete t[k];
                    after();
                };
                var undo = function () {
                    t[k] = v;
                    after();
                };
                hist.add({ Redo: redo, Undo: undo });
                return true;
            },
        });
    }

    var Hist = /** @class */ (function () {
        function Hist(depth) {
            this.depth_ = depth;
            this.list_ = [];
            this.offset_ = 0;
        }
        Object.defineProperty(Hist.prototype, "UndoLength", {
            get: function () { return this.list_.length - this.offset_; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hist.prototype, "RedoLength", {
            get: function () { return this.offset_; },
            enumerable: true,
            configurable: true
        });
        Hist.prototype.add = function (r) {
            var lst = this.list_.slice(0, this.list_.length - this.offset_);
            if (lst.length === this.depth_) {
                lst = lst.splice(0, 1);
            }
            if (lst.length === this.depth_)
                return;
            lst.push(r);
            this.offset_ = 0;
            this.list_ = lst;
            r.Redo();
        };
        Hist.prototype.undo = function () {
            if (this.list_.length) {
                var maxOffset = this.list_.length - 1;
                if (this.offset_ <= maxOffset) {
                    this.list_[maxOffset - this.offset_].Undo();
                    this.offset_++;
                }
            }
        };
        Hist.prototype.redo = function () {
            if (this.list_.length) {
                var maxOffset = this.list_.length - 1;
                if (this.offset_ > 0) {
                    this.offset_--;
                    this.list_[maxOffset - this.offset_].Redo();
                }
            }
        };
        Hist.prototype.clean = function () {
            this.list_ = [];
            this.offset_ = 0;
        };
        return Hist;
    }());

    /** Class of `hyperoop` top-level actions. When you properly set new values
     *  to entries of the `State` property then the corresponding DOM element will
     *  be redrawn automatically. The correspondence between `Actions` object
     *  and this DOM element is established by `init` function call.
     */
    var Actions = /** @class */ (function () {
        /** Construct an `Action` object, setting the initial `state` to it and optionally describing
         *  the `hist` object of type `redoundo.Hist`. If the `hist` argument is of the `number`
         *  type, the constructor will itself create the `History` object of the given length.
         */
        function Actions(start, hist) {
            this.orig = start;
            this.History = typeof hist === "number" ? new Hist(hist) : hist;
            this.init({ scheduleRender: function () { return null; } });
        }
        Object.defineProperty(Actions.prototype, "State", {
            /** This property, along with the `Remember` property, is used to change
             *  the state of an application and automatically redraw the user interface.
             */
            get: function () { return this.state; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Actions.prototype, "Remember", {
            /** This property has almost the same functionality as the `State` property,
             *  with the difference that changes made to it can be undone by calling the
             *  `History.undo` method.
             */
            get: function () { return this.remember; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Actions.prototype, "Renderer", {
            /** You can force redraw of the user interface by calling the `sheduleRender`
             *  method provided by this property. Usually you do not need to call it directly.
             */
            get: function () { return this.renderer; },
            enumerable: true,
            configurable: true
        });
        /** The page is redrawn automatically every time you set a new value for any entry of the `State`
         *  property. To set values for several inputs at once with only one automatic redraw, use
         *  this method, passing it a partial description of the new state and the optional parameter
         *  `remember` (`false` by default). If the value of the parameter `remember` is set to `true`, then
         *  the changes can be undone by calling `History.undo` method.
         */
        Actions.prototype.set = function (s, remember) {
            var _this = this;
            if (remember === void 0) { remember = false; }
            var keys;
            if (Array.isArray(s)) {
                keys = Array.from(s.keys());
            }
            else {
                keys = Object.getOwnPropertyNames(s);
            }
            keys = keys.filter(function (k) { return !(k in _this.orig) || _this.orig[k] !== s[k]; });
            var change = keys.length > 0;
            if (!change) {
                return;
            }
            var self = this;
            if (remember && this.History) {
                var was_1 = {};
                var wasnt_1 = [];
                for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
                    var k = keys_1[_i];
                    if (k in this.orig) {
                        was_1[k] = this.orig[k];
                    }
                    else {
                        wasnt_1.push(k);
                    }
                }
                this.History.add({
                    Redo: function () {
                        for (var _i = 0, keys_3 = keys; _i < keys_3.length; _i++) {
                            var k = keys_3[_i];
                            self.orig[k] = s[k];
                        }
                        self.renderer.scheduleRender();
                    },
                    Undo: function () {
                        for (var k in was_1) {
                            self.orig[k] = was_1[k];
                        }
                        for (var _i = 0, wasnt_2 = wasnt_1; _i < wasnt_2.length; _i++) {
                            var k = wasnt_2[_i];
                            delete self.orig[k];
                        }
                        self.renderer.scheduleRender();
                    },
                });
            }
            else {
                for (var _a = 0, keys_2 = keys; _a < keys_2.length; _a++) {
                    var k = keys_2[_a];
                    this.orig[k] = s[k];
                }
                this.renderer.scheduleRender();
            }
        };
        /** This method is used to initialize the `Actions` object by an instance of the `IRenderer` interface.
         *  Usually you do not need to call it directly.
         */
        Actions.prototype.init = function (r) {
            this.renderer = r;
            var self = this;
            this.state = make(this.orig, function () { return self.renderer.scheduleRender(); });
            this.remember = makeH(this.orig, function () { return self.renderer.scheduleRender(); }, this.History);
        };
        return Actions;
    }());
    /** Usually, you only need to manually initialize top-level `Action` objects (by calling the `init` function).
     * This class is used for lower level action-objects that inherit the necessary properties from higher level
     * objects. Thus, you do not need to initialize the `SubActions` of the lower levels manually.
     */
    var SubActions = /** @class */ (function (_super) {
        __extends(SubActions, _super);
        /** Construct an `Action` object, setting the initial `state` and parental `Actions`
         *  object. The last will be used for initializing new instance by renderer and history objects.
         */
        function SubActions(start, parent) {
            var _this = _super.call(this, start, parent.History) || this;
            if (parent.Renderer) {
                _this.init({ scheduleRender: function () { return parent.Renderer.scheduleRender(); } });
            }
            return _this;
        }
        return SubActions;
    }(Actions));

    function isVNode(n) {
        return !!n.nodeName;
    }
    function clone(target, source) {
        var out = {};
        for (var i in target) {
            out[i] = target[i];
        }
        for (var i in source) {
            out[i] = source[i];
        }
        return out;
    }
    function elementToVNode(element) {
        return {
            attributes: {},
            children: [].map.call(element.childNodes, function (el) { return el.nodeType === 3 // Node.TEXT_NODE
                ? el.nodeValue
                : elementToVNode(el); }),
            nodeName: element.nodeName.toLowerCase(),
        };
    }
    function resolveNode(node) {
        return typeof node === "function"
            ? resolveNode(node())
            : node != null
                ? node
                : "";
    }
    function getKey(node) {
        return node && isVNode(node) ? node.key : null;
    }
    function eventListener(event) {
        return event.currentTarget.events[event.type](event);
    }
    function updateStyle(element, value, oldValue) {
        if (typeof value === "string") {
            element.style.cssText = value;
        }
        else {
            if (typeof oldValue === "string") {
                oldValue = element.style.cssText = "";
            }
            for (var key in clone(oldValue, value)) {
                var style = value == null || value[key] == null ? "" : value[key];
                if (key[0] === "-") {
                    element.style.setProperty(key, style);
                }
                else {
                    element.style[key] = style;
                }
            }
        }
    }
    function updateEventProperty(element, name, value, oldValue) {
        name = name.slice(2);
        if (element.events) {
            if (!oldValue) {
                oldValue = element.events[name];
            }
        }
        else {
            element.events = {};
        }
        element.events[name] = value;
        if (value) {
            if (!oldValue) {
                element.addEventListener(name, eventListener);
            }
        }
        else {
            element.removeEventListener(name, eventListener);
        }
    }
    var specialAttrNames = {
        draggable: 1,
        list: 1,
        spellcheck: 1,
        translate: 1,
        type: 1,
    };
    function updateAttribute(element, name, value, oldValue, isSvg) {
        if (name === "key") {
            return;
        }
        if (name === "style") {
            updateStyle(element, value, oldValue);
        }
        else {
            if (name[0] === "o" && name[1] === "n") {
                updateEventProperty(element, name, value, oldValue);
            }
            else if (name in element && !(name in specialAttrNames) && !isSvg) {
                element[name] = value == null ? "" : value;
            }
            else if (value != null && value !== false) {
                element.setAttribute(name, value);
            }
            if (value == null || value === false) {
                element.removeAttribute(name);
            }
        }
    }
    function removeChildren(element, node) {
        if (isVNode(node)) {
            var attributes = node.attributes;
            for (var i = 0; i < node.children.length; i++) {
                removeChildren(element.childNodes[i], node.children[i]);
            }
            if (attributes.ondestroy) {
                attributes.ondestroy(element);
            }
        }
        return element;
    }
    function removeElement(parent, element, node) {
        function done() {
            parent.removeChild(removeChildren(element, node));
        }
        var cb = isVNode(node) && node.attributes && node.attributes.onremove;
        if (cb) {
            cb(element, done);
        }
        else {
            done();
        }
    }
    var Renderer = /** @class */ (function () {
        function Renderer(container, view, action) {
            this.skipRender = false;
            this.isRecycling = true;
            this.lifecycle = [];
            this.container = container;
            this.view = view;
            this.rootElement = (container && container.children[0]) || null;
            this.oldNode = this.rootElement && elementToVNode(this.rootElement);
            this.action = action;
        }
        Renderer.prototype.render = function () {
            var _this = this;
            this.skipRender = !this.skipRender;
            var node = resolveNode(function () {
                if (_this.action) {
                    _this.action.init(_this);
                }
                return _this.view();
            });
            if (this.container && !this.skipRender) {
                this.rootElement = this.patch(this.container, this.rootElement, this.oldNode, (this.oldNode = node));
            }
            this.isRecycling = false;
            while (this.lifecycle.length) {
                this.lifecycle.pop()();
            }
            return node;
        };
        Renderer.prototype.scheduleRender = function () {
            if (!this.skipRender) {
                this.skipRender = true;
                setTimeout(this.render.bind(this));
            }
        };
        Renderer.prototype.createElement = function (node, isSvg) {
            var element = null;
            if (typeof node === "string" || typeof node === "number") {
                element = document.createTextNode("" + node);
            }
            else {
                isSvg = isSvg || node.nodeName === "svg";
                if (isSvg) {
                    element = document.createElementNS("http://www.w3.org/2000/svg", node.nodeName);
                }
                else {
                    element = document.createElement(node.nodeName);
                }
            }
            if (isVNode(node)) {
                var attributes_1 = node.attributes;
                if (attributes_1) {
                    if (attributes_1.oncreate) {
                        this.lifecycle.push(function () { return attributes_1.oncreate(element); });
                    }
                    for (var i = 0; i < node.children.length; i++) {
                        element.appendChild(this.createElement((node.children[i] = resolveNode(node.children[i])), isSvg));
                    }
                    for (var name_1 in attributes_1) {
                        updateAttribute(element, name_1, attributes_1[name_1], null, isSvg);
                    }
                }
            }
            return element;
        };
        Renderer.prototype.updateElement = function (element, oldAttributes, attributes, isSvg) {
            for (var name_2 in clone(oldAttributes, attributes)) {
                var needUpdate = attributes[name_2] !==
                    (name_2 === "value" || name_2 === "checked" ? element[name_2] : oldAttributes[name_2]);
                if (needUpdate) {
                    updateAttribute(element, name_2, attributes[name_2], oldAttributes[name_2], isSvg);
                }
            }
            var cb = this.isRecycling ? attributes.oncreate : attributes.onupdate;
            if (cb) {
                this.lifecycle.push(function () { return cb(element, oldAttributes); });
            }
        };
        Renderer.prototype.patchNewNode = function (parent, element, oldNode, node, isSvg) {
            var newElement = this.createElement(node, isSvg);
            parent.insertBefore(newElement, element);
            if (oldNode != null) {
                removeElement(parent, element, oldNode);
            }
            return newElement;
        };
        Renderer.prototype.patchChildren = function (element, oldNode, node, isSvg) {
            if (isSvg === void 0) { isSvg = false; }
            isSvg = isSvg || node.nodeName === "svg";
            this.updateElement(element, oldNode.attributes, node.attributes, isSvg);
            var oldKeyed = {};
            var newKeyed = {};
            var oldElements = [];
            var oldChildren = oldNode.children;
            var children = node.children;
            for (var j = 0; j < oldChildren.length; j++) {
                oldElements[j] = element.childNodes[j];
                var oldKey = getKey(oldChildren[j]);
                if (oldKey != null) {
                    oldKeyed[oldKey] = [oldElements[j], oldChildren[j]];
                }
            }
            var i = 0;
            var k = 0;
            while (k < children.length) {
                var oldKey = getKey(oldChildren[i]);
                var newKey = getKey((children[k] = resolveNode(children[k])));
                if (newKeyed[oldKey]) {
                    i++;
                    continue;
                }
                if (newKey != null && newKey === getKey(oldChildren[i + 1])) {
                    if (oldKey == null) {
                        removeElement(element, oldElements[i], oldChildren[i]);
                    }
                    i++;
                    continue;
                }
                if (newKey == null || this.isRecycling) {
                    if (oldKey == null) {
                        this.patch(element, oldElements[i], oldChildren[i], children[k], isSvg);
                        k++;
                    }
                    i++;
                }
                else {
                    var keyedNode = oldKeyed[newKey] || [];
                    if (oldKey === newKey) {
                        this.patch(element, keyedNode[0], keyedNode[1], children[k], isSvg);
                        i++;
                    }
                    else if (keyedNode[0]) {
                        var el = element.insertBefore(keyedNode[0], oldElements[i]);
                        this.patch(element, el, keyedNode[1], children[k], isSvg);
                    }
                    else {
                        this.patch(element, oldElements[i], null, children[k], isSvg);
                    }
                    newKeyed[newKey] = children[k];
                    k++;
                }
            }
            while (i < oldChildren.length) {
                if (getKey(oldChildren[i]) == null) {
                    removeElement(element, oldElements[i], oldChildren[i]);
                }
                i++;
            }
            for (var j in oldKeyed) {
                if (!newKeyed[j]) {
                    removeElement(element, oldKeyed[j][0], oldKeyed[j][1]);
                }
            }
            return element;
        };
        Renderer.prototype.patch = function (parent, element, oldNode, node, isSvg) {
            if (isSvg === void 0) { isSvg = false; }
            if (node === oldNode) {
                return element;
            }
            if (oldNode == null || (oldNode.nodeName !== node.nodeName)) {
                element = this.patchNewNode(parent, element, oldNode, node, isSvg);
            }
            else if ((!isVNode(oldNode) || oldNode.nodeName == null) && !isVNode(node)) {
                element.nodeValue = "" + node;
            }
            else if (isVNode(node) && isVNode(oldNode)) {
                element = this.patchChildren(element, oldNode, node, isSvg);
            }
            else {
                throw new Error("Mom, what this 'patch' wants???");
            }
            return element;
        };
        return Renderer;
    }());

    /** Initialize DOM element `container` with virtual node `view` and optional
     *  `Actions` object `actions`. Calling this function is the only and necessary
     *  method of attaching a virtual tree to a DOM element.
     */
    function init(container, view, action) {
        var renderer = new Renderer(container, view, action);
        renderer.scheduleRender();
        return renderer;
    }

    exports.init = init;
    exports.h = h;
    exports.Actions = Actions;
    exports.SubActions = SubActions;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=hyperoop.umd.js.map
